import pytest
from mango_mdconverter import md2dict
from irods.meta import iRODSMeta


def test_simple_unflattening():
    metadict = {}
    md2dict.unflatten_namespace_into_dict(metadict, "simple_name", "simple_value")
    assert metadict == {"simple_name": "simple_value"}

    md2dict.unflatten_namespace_into_dict(metadict, "simple_name", "simple_value2")
    assert metadict == {"simple_name": ["simple_value", "simple_value2"]}

    md2dict.unflatten_namespace_into_dict(
        metadict, "second_name", "second_value", use_units=True
    )
    assert metadict == {
        "simple_name": ["simple_value", "simple_value2"],
        "second_name": "second_value",
    }

    md2dict.unflatten_namespace_into_dict(
        metadict, "third_name", "third_value", True, "units"
    )
    assert metadict == {
        "simple_name": ["simple_value", "simple_value2"],
        "second_name": "second_value",
        "third_name": ("third_value", "units"),
    }

    md2dict.unflatten_namespace_into_dict(
        metadict, "third_name", "fourth_value", units="units"
    )
    assert metadict == {
        "simple_name": ["simple_value", "simple_value2"],
        "second_name": "second_value",
        "third_name": [("third_value", "units"), "fourth_value"],
    }


def test_namespaced_unflattening():
    metadict = {}
    md2dict.unflatten_namespace_into_dict(metadict, "level1.level2.level3", "value")
    assert metadict == {"level1": {"level2": {"level3": "value"}}}

    md2dict.unflatten_namespace_into_dict(
        metadict, "level1.level2", "value2", True, "units"
    )
    assert metadict == {
        "level1": {"level2": {"level3": "value", "__value__": ("value2", "units")}}
    }


@pytest.fixture
def metadata_items():
    return [
        iRODSMeta("mgs.book.author.name", "Fulano De Tal", "1"),
        iRODSMeta("mgs.book.author.age", "50", "1"),
        iRODSMeta("mgs.book.author.pet", "cat", "1"),
        iRODSMeta("mgs.book.author.name", "Jane Doe", "2"),
        iRODSMeta("mgs.book.author.age", "29", "2"),
        iRODSMeta("mgs.book.author.pet", "cat", "2"),
        iRODSMeta("mgs.book.author.pet", "parrot", "2"),
        iRODSMeta("mgs.book.title", "A random book title"),
        iRODSMeta("mg.mime_type", "text/plain"),
        iRODSMeta("page_n", "567", "analysis/reading"),
        iRODSMeta("chapter_n", "15", "analysis/reading"),
    ]


@pytest.fixture
def bad_namespacing():
    return [
        iRODSMeta("ns.section", "1"),
        iRODSMeta("ns.section.title", "section_title"),
    ]


@pytest.fixture
def bad_namespacing_converted():
    return {"ns": {"section": {"title": "section_title", "__value__": "1"}}}


@pytest.fixture
def converted_dict():
    return {
        "schema": {
            "book": {
                "author": [
                    {"age": "50", "name": "Fulano De Tal", "pet": "cat"},
                    {"age": "29", "name": "Jane Doe", "pet": ["cat", "parrot"]},
                ],
                "title": "A random book title",
            }
        },
        "mg": {"mime_type": "text/plain"},
        "analysis": {"reading": {"page_n": "567", "chapter_n": "15"}},
    }


def test_unpacking(metadata_items):
    metadict = {}
    md2dict.unpack_metadata_into_dict(metadict, metadata_items[0])
    assert metadict == {"mgs": {"book": {"author": {"name": ("Fulano De Tal", "1")}}}}

    md2dict.unpack_metadata_into_dict(metadict, metadata_items[1])
    assert metadict == {
        "mgs": {
            "book": {"author": {"name": ("Fulano De Tal", "1"), "age": ("50", "1")}}
        }
    }

    md2dict.unpack_metadata_into_dict(
        metadict, iRODSMeta("chapter_n", "15", "analysis/reading")
    )
    assert metadict == {
        "mgs": {
            "book": {"author": {"name": ("Fulano De Tal", "1"), "age": ("50", "1")}}
        },
        "chapter_n": ("15", "analysis/reading"),
    }


def test_mango_conversion(metadata_items, converted_dict):
    reorganized_dict = md2dict.convert_metadata_to_dict(metadata_items)
    assert reorganized_dict == converted_dict

    # metadata_items.reverse()
    # reorganized_dict2 = md2dict.convert_metadata_to_dict(metadata_items)
    # assert reorganized_dict2 == converted_dict


def test_bad_namespacing_conversion(bad_namespacing, bad_namespacing_converted):
    reorganized_dict = md2dict.convert_metadata_to_dict(bad_namespacing)
    assert reorganized_dict == bad_namespacing_converted

    bad_namespacing.reverse()
    reorganized_dict = md2dict.convert_metadata_to_dict(bad_namespacing)
    assert reorganized_dict == bad_namespacing_converted



from irods.meta import iRODSMeta
from mango_mdschema.helpers import flattened_from_mango_avu, unflatten


def safely_add_to_dict(regular_dict: dict, key, value, units=None):
    """Compile dictionary with key-value pairs or key-(value, units) pairs from AVUs.
    Make lists out of multivalued fields."""
    _value = value if units is None else (value, units)
    if key in regular_dict:
        if type(regular_dict[key]) == list:
            regular_dict[key].append(_value)
        elif type(regular_dict[key]) == dict and type(_value) != dict:
            safely_add_to_dict(regular_dict[key], "__value__", value, units)
        elif (existing_value := regular_dict[key]) is not None:
            regular_dict[key] = [existing_value, _value]
        else:  # basically None value
            regular_dict[key] = _value
    else:
        regular_dict[key] = _value


def unflatten_namespace_into_dict(
    target_dict: dict,
    namespaced_string: str,
    value=None,
    use_units=False,
    units=None,
) -> dict:
    """Expand namespaced string (with dots) into a nested dict

    Args:
        target_dict (dict): Dictionary to update: expanded AVU names
          will be keys, and values or tuples of values and units will
          be values.
        namespaced_string (str): AVU name
        value (str, optional): AVU value. Defaults to None.
        use_units (bool, optional): Whether units should be included. Defaults to False.
        units (str, optional): AVU units. Defaults to None.

    Returns:
        dict: Updated dictionary
    """
    if "." in namespaced_string:
        lead_key, rest = namespaced_string.split(".", 1)
        if lead_key not in target_dict:
            target_dict[lead_key] = {}
        if not isinstance(target_dict[lead_key], dict):
            target_dict[lead_key] = {"__value__": target_dict[lead_key]}
        unflatten_namespace_into_dict(
            target_dict[lead_key], rest, value, use_units, units
        )
    else:
        safely_add_to_dict(
            target_dict, namespaced_string, value, units if use_units else None
        )


def unpack_metadata_into_dict(target_dict: dict, avu: iRODSMeta) -> dict:
    """From a list of AVUs to a dictionary

    Example:
        metadict = {}
        for avu in obj.metadata.items():
            unpack_metadata_into_dict(metadict, avu)
    """
    unflatten_namespace_into_dict(target_dict, avu.name, avu.value, True, avu.units)


def prepare_metadata_for_download(metadict: dict, no_label: str = "other") -> dict:
    """Reorganize nested dictionary with schemas and analysis AVUs separated.
    To run on the output of `unpack_metadata_into_dict()`"""

    reorganized_metadata = {}
    for k, v in metadict.items():
        if type(v) == dict:
            if k == "schema":
                for schema_name, schema_avus in v.items():
                    v[schema_name] = unflatten(
                        list(
                            map(
                                lambda x: flattened_from_mango_avu(
                                    x, f"mgs.{schema_name}"
                                ),
                                schema_avus,
                            )
                        )
                    )
            reorganized_metadata[k] = v
        elif type(v) == tuple and v[1].startswith("analysis/"):
            if not "analysis" in reorganized_metadata:
                reorganized_metadata["analysis"] = {}
            analysis_subtype = v[1].replace("analysis/", "")
            if not analysis_subtype in reorganized_metadata["analysis"]:
                reorganized_metadata["analysis"][analysis_subtype] = {}
            reorganized_metadata["analysis"][analysis_subtype][k] = v[0]
        else:

            def remove_units(value):
                return value[0] if type(value) == tuple else value

            if not no_label in reorganized_metadata:
                reorganized_metadata[no_label] = {}
            reorganized_metadata[no_label][k] = (
                remove_units(v)
                if type(v) != list
                else [remove_units(value) for value in v]
            )
    return reorganized_metadata


def convert_metadata_to_dict(metadata_items) -> dict:
    """Convert iterable of iRODSMeta into nested dictionary

    Args:
        metadata_items (list of iRODSMeta): Metadata items from one data object or collection

    Example:
        convert_metadata_to_dict(irods_item.metadata.items())

    """
    metadict = {"schema": {}}
    for item in metadata_items:
        if item.name.startswith("mgs"):
            schema_name = item.name.split(".")[1]
            if schema_name not in metadict["schema"]:
                metadict["schema"][schema_name] = []
            if item.name.count(".") > 2 and item.units is None:
                item.units = ".".join(["1"] * (item.name.count(".") - 2))
            metadict["schema"][schema_name].append(item)
        else:
            unpack_metadata_into_dict(metadict, item)
    if len(metadict["schema"]) == 0:
        del metadict["schema"]
    return prepare_metadata_for_download(metadict)


def filter_metadata_dict(metadict, filters: dict = {}) -> dict:
    """Filter an dictionary of metadata to only retrieve certain keys.

    Args:
        metadict (any): Initially, the metadata dictionary, but then
          it is applied recursively to any value within the metadata dictionary.
        filters (dict, optional): A dictionary of keys, indicating which keys of
          the metadict should be included. For nested fields, the value should be
          another dict of the same format. Otherwise, an empty dictionary or
          `None` as value is enough. Defaults to `{}`, which results in the
          full dictionary being returned.

    Returns:
        dict: A (filtered) dictionary of metadata
    """
    if filters is None or len(filters) == 0:
        return metadict
    if not isinstance(filters, dict):
        raise TypeError("The 'filters' argument should be a dictionary or empty.")
    if isinstance(metadict, list):
        return [filter_metadata_dict(d, filters) for d in metadict]
    if isinstance(metadict, dict):
        return {
            k: filter_metadata_dict(v, filters.get(k, {}))
            for k, v in metadict.items()
            if k in filters
        }
    return metadict
